:- module(
  commandline,
  [
    do_arguments/0
  ]
).

:- doc('
\\section{Biocham files}

Biocham file names are suffixed by \\texttt{.bc}. In a Biocham file, everything following the character percent is a comment.
Some other file formats are used.').

:- doc('
Biocham models can be imported from, and exported to, other file formats using the following suffixes:
\\begin{itemize}
\\item \\texttt{.xml} for Systems Biology Markup Language (SBML) files, more precisely SBML2 files for reaction networks and SBML3qual files for influence networks;
\\item \\texttt{.ode} for Ordinary Differential Equation files (ODEs in XXPAUT format), allowing us to infer and import a reaction network from ODEs using a heuristic inference algorithm described in \\cite{FGS15tcs}.
\\item \\texttt{.ipynb} for Jupyter notebook files.
\\end{itemize}
Biocham numerical data time series can be imported/exported in 
\\begin{itemize}
\\item \\texttt{.csv} comma-separated values format (spreadsheet format).
\\end{itemize}
The following files can also be used to export some Biocham objects:
\\begin{itemize}
\\item \\texttt{.tex} LaTeX format for exporting ODEs and graphs;
\\item \\texttt{.dot} for exporting graphs;
\\item \\texttt{.plot} for exporting numerical data time series;
\\item \\texttt{.smv} for exporting boolean transition systems and Computation Tree Logic (CTL) queries for the NuSMV model-checker;
\\item \\texttt{.dot} for exporting graphs.
\\end{itemize}
').

:- doc('\\section{Biocham call options}').

do_arguments :-
  current_prolog_flag(argv, Argv),
  do_arguments(Argv).


:- dynamic(file/1).


do_arguments(Argv) :-
  retractall(file(_)),
  parse_arguments(Argv),
  load_files.


load_files :-
  \+ (
    file(File),
    \+ (
      load(File)
    )
  ).


parse_arguments([]).

parse_arguments([Arg | Argv]) :-
  (
    Arg = '--'
  ->
    add_files(Argv)
  ;
    atom_concat('--', _, Arg)
  ->
    (
      option(_, Arg, Argv, ArgTail)
    ->
      parse_arguments(ArgTail)
    ;
      throw(error(unknown_option(Arg)))
    )
  ;
    add_file(Arg),
    parse_arguments(Argv)
  ).


option('', '--trace', Arg, Arg) :-
  leash(-all),
  trace.

option('', '--generate-doc', Arg, Arg) :-
  generate_doc,
  quit.

option('', '--version', Arg, Arg) :-
   quit.

option('', '--jupyter', [], []) :-
  set_counter(graph_png, 0),
  set_counter(plot_csv, 0),
  set_counter(plot_png, 0),
  set_plot_driver(gnu_plot_csv),
  set_image_viewer_driver(img_tag),
  set_draw_graph_driver(graph_png).

option('', '--list-commands', [], []) :-
  forall(
    (
      current_functor(Functor, Arity),
      predicate_info(Functor/Arity, _ArgumentTypes, _, yes, _)
    ),
    (
      % Pred =.. [Functor | ArgumentTypes],
      format('~a~n', [Functor])
    )
  ),
  quit.

:- doc('\\texttt{biocham [file]}
launches Biocham and optionnally loads the file given as argument.').

:- doc('\\texttt{biocham --trace --generate-doc --version --list-commands} are possible options (prefixed by two dashes).').

:- doc('\\texttt{biocham --notebook [notebook.ipynb]}  launches Jupyter (see
  \\url{http://jupyter-notebook.readthedocs.io/}) with a biocham kernel, and
  loads the corresponding notebook (ipynb suffix). If no notebook is provided, a biocham notebook can be created with the new menu.

To execute a Biocham command enter Shift-Return.

All shortcuts are described in the \\emph{keyboard menu}.
 The following video
  \\url{https://www.youtube.com/watch?v=e9cSF3eVQv0} is a quite nice
  introduction to the Jupyter notebook''s concepts, all can be adapted to
  Biocham as a kernel.
  \\html{<div><iframe width="640" height="360"
  src="https://www.youtube-nocookie.com/embed/e9cSF3eVQv0?rel=0&start=287&end=779"
  frameborder="0" allowfullscreen></iframe></div>}

  The only \\emph{magic commands} available in the notebook and not in Biocham are 

\\texttt{%load file.bc} which creates a cell for each command contained in a Biocham file

and
  \\texttt{%slider k1 k2 …}, which  creates sliders to change the given
  parameters'' value, and run \\command{numerical_simulation} followed by
  \\command{plot} at each change.
').

:- doc('
  The environment variable \\texttt{BIOCHAM_TIMEOUT} can be used to change the
  default notebook-kernel communication timeout of 120 (seconds).
').

:- doc('\\texttt{biocham_debug} launches Biocham with the Prolog toplevel.').

option('', '--notebook', Args, []) :-
  Cmd =.. [
    jupyter,
    notebook,
    '--KernelSpecManager.whitelist=["biocham"]',
    '--NbConvertBase.display_data_priority=["image/png", "text/html"]'
    | Args
  ],
  library_path(Lib),
  file_directory_name(Lib, Dir),
  % Change to biocham directory to be sure to find the biocham_kernel Python
  % module
  (
    getenv('PYTHONPATH', Pythonpath)
  ->
    true
  ;
    Pythonpath = ''
  ),
  atomic_list_concat([Pythonpath, Dir], ':', NewPythonpath),
  setenv('PYTHONPATH', NewPythonpath),
  exec(Cmd),
  quit.


add_files(Files) :-
  \+ (
    member(File, Files),
    \+ (
      add_file(File)
    )
  ).


add_file(File) :-
  assertz(file(File)).
