#include <stdlib.h>
#include <cstddef>
#include <ppl.hh>
#include "ppl_validity_domain.hh"
extern "C" {
#include "gsl_solver.h"
}
#include "cmaes_interface.h"
#include "satisfaction_degree.hh"
#include "ode.inc"
#include "check.inc"
#include "search_parameters.inc"
#include "domain_distance.inc"

static double
fitfun(gsl_solver *solver, gsl_solver_config *config, double values[DIMENSION])
{
    gsl_solver_state state;
    int i;
    double *p = solver->p;
    Table table;
    double penalty = 0;
    init_gsl_solver_state(&state, solver);
    // reset step size and evolution, forces again initial step size because
    // otherwise seems useless
    gsl_odeiv2_driver_reset_hstart(solver->d, config->initial_step_size);
    for (i = 0; i < DIMENSION; i++) {
        const struct search_parameter *param = &search_parameters[i];
        double value = values[i];
        p[param->index] = value;
        if (value < param->min) {
            penalty += param->min - value;
        }
        else if (value > param->max) {
            penalty += value - param->max;
        }
    }
    /* char filename[30]; */
    /* sprintf(filename, "ode_%.17g.csv", values[0]); */
    /* FILE *csv = fopen(filename, "w"); */
    /* print_headers(csv); */
    /* i = 0; */
    do {
        Row row;
        add_row(row, state);
        table.push_back(row);
        /* print(csv, &state); */
        /* i++; */
    } while (gsl_solver_step(&state));
    free_gsl_solver_state(&state);
    PPL::set_rounding_for_PPL();
    Domain domain = compute_domain(table);
    /* using PPL::IO_Operators::operator<<; */
    /* std::cerr << domain << "." << std::endl; */
    /* std::cerr << i << " rows" << std::endl; */
    double result = domain_distance(domain);
    PPL::restore_pre_PPL_rounding();
    if (penalty > 0) {
        penalty += 10000;
    }
    return penalty + result;
}


int
main(int argc, char *argv[])
{
    cmaes_t evo;
    double *arFunvals, *const*pop, *xfinal;
    double xstart[DIMENSION], stddev[DIMENSION];
    int i;
    /* int j; */
    gsl_solver_config config;
    gsl_solver solver;
    bool first;
    double best_vd = 1;
    init_gsl_solver_config(&config);
    init_gsl_solver(&solver, &config);
    for (i = 0; i < DIMENSION; i++) {
        const struct search_parameter *param = &search_parameters[i];
        double min = param->min;
        double max = param->max;
        xstart[i] = param->init;
        stddev[i] = (max - min) / 2;
    }
    arFunvals = cmaes_init(
        &evo, DIMENSION, xstart, stddev, SEED, 0, "cmaes_initials.par");
    cmaes_ReadSignals(&evo, "cmaes_signals.par");
    while(!cmaes_TestForTermination(&evo) && best_vd >= 0.0001) {
        pop = cmaes_SamplePopulation(&evo);
        for (i = 0; i < cmaes_Get(&evo, "lambda"); ++i) {
            arFunvals[i] = fitfun(&solver, &config, pop[i]);
            /* fprintf(stderr, "distance: %f\n", arFunvals[i]); */
            /* for (j = 0; j < DIMENSION; ++j) { */
            /*     fprintf(stderr, "   %.17g\n", pop[i][j]); */
            /* } */
        }
        cmaes_UpdateDistribution(&evo, arFunvals);
        best_vd = cmaes_Get(&evo, "fbestever");
        /* fprintf(stderr, "Best satisfaction degree: %f\n", */
        /*         1 / (1 + best_vd)); */
        /* xfinal = cmaes_GetNew(&evo, "xbestever"); */
        /* for (i = 0; i < DIMENSION; i++) { */
        /*     fprintf(stderr, "   %f\n", xfinal[i]); */
        /* } */
        cmaes_ReadSignals(&evo, "cmaes_signals.par");
        fflush(stdout);
    }
    cmaes_WriteToFile(&evo, "all", "allcmaes.dat");
    xfinal = cmaes_GetNew(&evo, "xbestever");
    for (i = 0; i < DIMENSION; i++) {
        /* fprintf(stderr, "%.17g\n", xfinal[i]); */
        printf("%.17g\n", xfinal[i]);
    }
    printf("%f\n", 1 / (1 + best_vd));
    cmaes_exit(&evo);
    free(xfinal);
    free_gsl_solver(&solver);
    return EXIT_SUCCESS;
}
